/*
 * Reciprocal frequency counter with LCD output
 *
 * (c) 2010 Matthias Hopf <mat@mshopf.de>
 * License: GPL v2
 */

#include "cpu_mhz.h"

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/eeprom.h>
#include <util/delay.h>

#include <stdint.h>
#include <stdlib.h>
#include <math.h>


#define INIT_TEXT       "FreqC V1.0"
#define MIN_SAMPLE_TIME	0.5	// May be extended by up to a full cycle
#define MIN_RES		5	// Minimum resolution: 6 numbers  i.e. .000001
#define MAX_RES		7	// Maximum resolution: 7 numbers  i.e. 123456.7

#define DEBUG_LED	0	// Enable to have blinking lights
#define DEBUG_LCD	0	// Only working with min. 2x8 displays

#if DEBUG_LCD
#  define MAX_RES	0	// Required to get in 2k
#endif


/*
 * LCD definitions
 */

#define LCD_LINES	2	// 1 or 2 lines
#define LCD_COLS	5	// 10col display (2*5) - only required for wrap

#define LCD_PORTNAME	B
#define LCD_D4		0
#define LCD_RS		6
#define LCD_RW		5
#define LCD_E		4

#define BIND static
#include "lcd_hd44780u.h"


/*
 * 64bit calc ops
 */

#define BIND static
#include "uint64_ops.h"


/*
 * Other port pins
 */

#if DEBUG_LED			// enable to have blinking lights
#  define LED_PORT	PORTD
#  define LED_DDR	DDRD
#  define LED_BITS	(1<<5)
#endif


/*
 * Helper routines
 */

/* Setup decimal point on a fixpoint number representation */
static inline void buf_fixpoint (char *buf, uint8_t prec)
{
    uint8_t i, len = 0;

    while (buf[len])		// strlen()
	len++;

    if (len <= prec) {
	i = len+1;
	while (i--)
	    buf[i + prec - len + 1] = buf[i];
	*buf++ = '.';
	while (len++ < prec)
	    *buf++ = '0';
    } else {
	i = len+1;
	while (i-- >= len - prec)
	    buf[i + 1] = buf[i];
	buf[len - prec] = '.';
    }
}


/*
 * Globals
 */

/* Counter 0:   Counting incoming events. 8bit + 32bit software */
/* Max. sampling speed F_CPU/2 = 10MHz, at 1 second with 8bit timer this would
 * make a maximum overflow of 39062 and enough for 16bit.
 * However, we want longer potential sampling rates, and the next falling edge
 * could be further away, thus 32bit. */
volatile uint32_t cnt_t0;
/* Counter 1:   Counting system clock. 16bit counter + 16bit software */
/* Counter is 16bit wide, thus 16bit overflow is enough. */
volatile uint16_t cnt_t1;

/*
 * Interrupts
 */

/* Interrupt on falling edge for starting / stopping timer interrupts */
ISR(INT0_vect)
{
    /* Luckily, this timer only has to toggle timer interrupts
     * - thus no distinction of cases required here. */
    /* Enable / disable both timer */
    TCCR0B ^= 0x06 << CS00;		// T0 falling edge
    TCCR1B ^= 0x01 << CS10;		// clk/1

    /* Disable edge interrupt - don't get called again */
    GIMSK = 0x00;
    EIFR  = 0x40;	// clear potential spurious interrupt triggered again flag
}

/* Overflow interrupt for event counter */
ISR(TIMER0_OVF_vect)
{
    cnt_t0++;
}

/* Overflow interrupt for system clock counter */
ISR(TIMER1_OVF_vect)
{
    cnt_t1++;
    /* Check for sampling time */
    if (cnt_t1 == (uint16_t) (MIN_SAMPLE_TIME * F_CPU / 65536.0)) {
	/* Re-enable external edge triggering interrupt for disabling timers */
	GIMSK = 0x40;
	EIFR  = 0x40;
	/* Worst case for precision: falling edge now, on a signal that is
	 * *almost* fast enough to trigger a second count after this
	 * instruction before INT0 kicks in and can disable the counters */
	/* TODO: improve precision by using assembler here */
    }
}


/*
 * Main
 */
__attribute__((naked))		// Suppress redundant SP initialization
extern int main (void)
{
    uint8_t i;
    static char buf[16];	// Fails if non-static?!?

#if DEBUG_LED
    /* Initial blinking to show we're alive */
    LED_DDR |= LED_BITS;
    for (i = 0; i < 10; i++) {
	_delay_ms (100.0);
	LED_PORT ^= LED_BITS;
    }
#endif

    /* Init LCD */
    lcd_init     ();
    _delay_ms    (200.0);
    lcd_putswrap (0, 0, INIT_TEXT);
    _delay_ms    (500.0);

    /* Setup pull-up for external input - allow open collector inputs */
    /* INT0 and T0 have to be connected */
    /* TODO: probably different on mega */
    DDRD  &= ~(1 << 2);		// Pull-up IN PD2=INT0  PD4=T0 connected
    PORTD |=  (1 << 2);
    MCUCR = 0x02;		// INT0 external port interrupts on falling edge

    /* Setup counters */
    /* No compare, no waveform, normal mode, off */
    TCCR0A = TCCR1A = 0;
    TCCR0B = TCCR1B = 0;
#if defined (__AVR_ATtiny2313__) || defined (__AVR_ATtiny2313A__)
    TIMSK  = 0x82;		// TOIE0 Cnt0 Overflow Interrupt | TOIE1
#else // mega
    /* TODO: untested */
    TIMSK0 = 0x01;		// TOIE0 Cnt0 Overflow Interrupt
    TIMSK1 = 0x01;		// TOIE1
#endif

    /*
     * Main loop
     */
    for (;;)
    {
	/*
	 * Start measurement
	 */
//	lcd_cmd (0x0f);		// Enable cursor blink
	lcd_cmd (0x0e);		// Enable cursor on
	lcd_pos (1, 2);		// Cursor at 1,2 while waiting for first edge

	/* Clear counters */
	TCNT0  = 0;
	TCNT1  = 0;
	cnt_t0 = 0;
	cnt_t1 = 0;

	/* Enable port0 interrupts */
	GIMSK = 0x40;
	EIFR  = 0x40;		// clear old interrupt flag
	sei();

	/* Wait until falling edge on INT0 enabled timer interrupts */
	while (! TCCR0B)
	    ;

	while (TCCR0B)
	{
#if DEBUG_LCD
	    uint32_t u32;
	    /* Output event count on 1,[0-3] - overflow counter only if high freq */
	    cli();
	    u32 = cnt_t0;
	    sei();
	    if (u32) {			// High count at 1,1
		ultoa (u32, buf, 10);
		lcd_pos  (1, 1);
		lcd_nputs (buf, 4);
	    } else {			// Low count at 1,0
		lcd_pos  (1, 0);
		utoa (TCNT0, buf, 10);
		lcd_nputs (buf, 4);
	    }

	    /* Output overflow clock count on 1,[5-7] */
	    lcd_pos (1, 5);
	    cli();
	    u32 = cnt_t1;
	    sei();
	    ultoa (u32, buf, 10);
	    lcd_nputs (buf, 3);
#endif

	    lcd_pos (1, 3);		// Cursor at 1,3 while sampling
	    if (GIMSK)
		lcd_pos (1, 4);		// Cursor at 1,4 while waiting for last edge
	    _delay_ms (100.0);
#if DEBUG_LED
	    LED_PORT ^= LED_BITS;
#endif
	}

	/* Make sure that potential left-over t0 ovf interrupt is handled */
	__asm__ __volatile__ ("nop");
	/* before disabling interrupts */
	cli();

	/*
	 * Calculate frequency from counters
	 */
	uint32_t cnt = TCNT0 + ((uint32_t)cnt_t0 << 8);
	const uint64_t fact = F_CPU * pow (10, MIN_RES) + .5;
	/* This would overflow in 32bit
	 * (actually fact may even be larger than UINT32_MAX) */
	uint64_t frq = uint64_mul32 (fact, cnt);
	uint32_t nom = TCNT1 + ((uint32_t)cnt_t1 << 16);
	frq += nom >> 1;		// round
	frq = uint64_div32  (frq, nom);

	/* Find precision (auto-range) */
	uint8_t s = MIN_RES;
#if MAX_RES > 0
	while (s && frq >= (uint64_t) pow (10, MAX_RES)) {
	    frq += 5;			// round
	    frq = uint64_div32  (frq, 10);
	    s--;
	}
#endif

	/* Format */
	ultoa ((uint32_t)frq, buf, 10);
	if (s)
	    buf_fixpoint (buf, s);

	/*
	 * Print to LCD
	 */
	lcd_cmd   (0x0c);
	lcd_clear ();
#if LCD_COLS > 5
	lcd_puts  (buf);
	lcd_pos   (1, 0);
	lcd_puts  ("Hz");
#else
	for (i = 0; buf[i]; i++)
	    ;
	buf[i++] = 'H';
	buf[i++] = 'z';
	buf[i]   = 0;
	lcd_putswrap (0, 0, buf);
#endif
    }
}
